//=============================================================================
// File: diskpartition.cpp
//    
// Created Thu Apr 13 11:53:11 2000
//    
// Copyright (c) 2000 
// Justin E. Forrester (jforrest@cs.wisc.edu)
// Jonathan Ledlie (ledlie@cs.wisc.edu)
// All Rights Reserved
// 
// Description:
//    impl of disk part methods 
//     
//=============================================================================
#include "diskpartition.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <iostream>
#include <unistd.h>


//=============================================================================
// Method: ctor
//     Description:
//         constructor
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
DiskPartition::DiskPartition (void) 
{
    fd = 0;
    tupleCount = 0;
    memset (fileName, 0, MAX_FILENAME);  
    tsList = NULL;
} // end ctor


//=============================================================================
// Method: dtor
//     Description:
//         destructor
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
DiskPartition::~DiskPartition (void) 
{

} // end dtor


//=============================================================================
// Method: cleanup
//     Description:
//         clean up the memory from the disk partition, and delete the file
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::cleanup (void) 
{
  // delete all of our timestamplistnodes
  TimestampListNode *current = tsList;
  TimestampListNode *previous;
  while (current != NULL) {
    previous = current->next;
    delete current;
    current = previous;
  }
  unlink (fileName);
  return OK;

} // end cleanup


//=============================================================================
// Method: setFilename
//     Description:
//         set the filename of the backing store for this disk partition
//     Parameters:
//         name - filename
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition:: DiskPartition::setFilename(char *name) 
{
    strcpy (fileName, name);
    return OK;
} // end setFilename


//=============================================================================
// Method: beginAppend
//     Description:
//         start the append process (opening file)
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::beginAppend(void) 
{
    fd = open (fileName, O_WRONLY|O_CREAT|O_APPEND, S_IWUSR | S_IRUSR);
    if (fd > 0)
        return OK;
    return FILE_ERROR;

} // end beginAppend


//=============================================================================
// Method: appendTuple
//     Description:
//         write one tuple to the disk partition
//     Parameters:
//         st - tuple to write
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::appendTuple(StampedTuple *st) 
{
    // set DTS for the tuple
    st->DTS.setToCurrTime();

    // update the current last DTS for the list entries
    lastDTS = st->DTS;

    // write out the two timestamps
    int bytesWritten = write (fd, st, sizeof (Timestamp) * 2);
    if (bytesWritten <= 0)
    {
        return FILE_ERROR;
    }
    // write out the tuple itself
    bytesWritten = write (fd, st->tuple, sizeof (WITuple));
    if (bytesWritten <= 0)
    {
        return FILE_ERROR;
    }

    // update the count of tuples on disk
    tupleCount++;

    return OK;
} // end appendTuple


//=============================================================================
// Method: endAppend
//     Description:
//         end the current appending process (closes file)
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::endAppend(void) 
{
    if (close (fd) == 0) 
        return OK;
    return FILE_ERROR;

} // end endAppend


//=============================================================================
// Method: beginRead
//     Description:
//         start reading from backing store (open file)
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::beginRead(void) {

    fd = open (fileName, O_RDONLY);
    if (fd > 0)
        return OK;
    return FILE_ERROR;
} // end beginRead


//=============================================================================
// Method: nextTuple
//     Description:
//         Read in the next tuple and copy it into the reference to the 
//         StampedTuple argument
//     Parameters:
//         st - stamped tuple to copy data into
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::nextTuple(StampedTuple *st) 
{
    int bytesRead = read (fd, st, sizeof (Timestamp) * 2);
    bytesRead += read (fd, st->tuple, sizeof (WITuple));
    if (bytesRead != (sizeof(Timestamp)*2 + sizeof(WITuple)))
    {
        return FILE_ERROR;
    }
   
    return OK;

} // end nextTuple


//=============================================================================
// Method: endRead
//     Description:
//         end current read process (close file)
//     Parameters:
//         addListNode - if true, add 2nd stage activation record
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::endRead(bool addListNode) 
{

    if (addListNode)
    {
        // add a new node because this signifies the beginning of the 2nd stage
        TimestampListNode *pNewNode = new TimestampListNode();
        pNewNode->lastDTS = this->lastDTS;
        pNewNode->probeTS.setToCurrTime();
        pNewNode->next = tsList;
        tsList = pNewNode;
    }

    if (close (fd) == 0) 
        return OK;
    return FILE_ERROR;
}


//=============================================================================
// Method: checkDups
//     Description:
//         check for duplicates using 2nd stage activation list
//     Parameters:
//         tupleFromThisDP - tuple being joined from this disk partition
//         mathingTuple - tuple being joined from other relation mem partition
//     Return Value:
//         error status
//=============================================================================
Status DiskPartition::checkDups(StampedTuple *tupleFromThisDP,
                                StampedTuple *matchingTuple)
{
    // iterate through the list of ts's
    TimestampListNode *pCurrTSNode = tsList;
    while (pCurrTSNode != NULL)
    {
        if (tupleFromThisDP->DTS <= pCurrTSNode->lastDTS &&
            matchingTuple->ATS < pCurrTSNode->probeTS &&
            matchingTuple->DTS > pCurrTSNode->probeTS)
        {
            // this is a duplicate
            return ERR_DUPLICATE;
        }
        pCurrTSNode = pCurrTSNode->next;
    }
    return OK;
} // end checkDups


//=============================================================================
// Method: getTupleCount
//     Description:
//         get current number of tuples residing in disk partition
//     Parameters:
//         none
//     Return Value:
//         number of tuples
//=============================================================================
int DiskPartition::getTupleCount(void)
{
    return tupleCount;
}


//=============================================================================
// Method: operator<<
//     Description:
//         useful for debugging
//     Parameters:
//         -
//     Return Value:
//         -
//=============================================================================
ostream& operator << (ostream& os, DiskPartition& diskPartition) 
{
    os << "disk (" << diskPartition.tupleCount << ")";
    return os;
}
