//=============================================================================
// File: mempartition.cpp
//    
// Created Tue Apr 11 13:23:46 2000
//    
// Copyright (c) 2000 
// Justin E. Forrester (jforrest@cs.wisc.edu)
// Jonathan Ledlie (ledlie@cs.wisc.edu)
// All Rights Reserved
// 
// Description:
//    implementation of mem partition 
//     
//=============================================================================
#include <iostream.h>
#include "mempartition.h"
#include "output.h"

int memIdCount = 0;

//=============================================================================
// Method: ctor
//     Description:
//         constructor
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
MemPartition::MemPartition (void) 
{
    tupleCount = 0;
    id = memIdCount++;

    for (int i = 0; i < NUM_CHAINS; i++)
    {
        tuples[i] = NULL;
    }
} // end ctor


//=============================================================================
// Method: dtor
//     Description:
//         destructor
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
MemPartition::~MemPartition (void) 
{

} // end dtor


//=============================================================================
// Method: finalMemProbe
//     Description:
//         perform final mem probe to catch any matches on items that 
//         were never flushed to disk
//     Parameters:
//         probeeMemPart - memory partition that we will probe with this 
//                         partition's tuples
//     Return Value:
//         error status
//=============================================================================
Status MemPartition::finalMemProbe(MemPartition *probeeMemPart)
{
    int freeCount = 0;
    for (int i = 0; i < NUM_CHAINS; i++) 
    {
        StampedTuple *currTuple = tuples[i];
        StampedTuple *previous;
        while (currTuple != NULL) 
        {
            previous = currTuple->next;
            StampedTuple *pMatchingTuple = NULL;
            // check to see if they join
            if (OK == probeeMemPart->probe(currTuple, pMatchingTuple))
            {
                // check to see if they were joined in the first stage
                if (!(currTuple->ATS > pMatchingTuple->ATS &&
                      currTuple->ATS < pMatchingTuple->DTS) &&
                    !(pMatchingTuple->ATS > currTuple->ATS &&
                      pMatchingTuple->ATS < currTuple->DTS))
                {
                    // we did not match these in the first stage
                    
                    // check to see if we already matched these in another
                    // run of the second stage
                    if (OK == probeeMemPart->getDiskPartition()->checkDups(pMatchingTuple, currTuple))
                    {
                        Timestamp max;
                        max.setToMaxTime();
                        
                        ASSERT(currTuple->DTS == max);

                        // we have a legitimate match
                        OutputTuple *output = new OutputTuple;
                        if ((output->tuple1.tuple = new WITuple) == 0)
                            perror("error allocating");
                        output->tuple2.tuple = new WITuple;
                        
                        // copy the data into it
                        memcpy (output->tuple1.tuple, currTuple->tuple, sizeof(WITuple));
                        memcpy (output->tuple2.tuple, pMatchingTuple->tuple, sizeof(WITuple));
                        
                        // send it to the output object
                        Output::writeTuple (output, false);
                    }
                }
            }
            pMatchingTuple = NULL;
        
            freeCount++;
            delete currTuple->tuple;
            delete currTuple;
            currTuple = previous;
        }
        tuples[i] = NULL;
    }
    return OK;

} // end finalMemProbe


//=============================================================================
// Method: setDiskPartition
//     Description:
//         set the corresponding disk parition
//     Parameters:
//         dp - disk partition that this mem partition will go to when flushed
//     Return Value:
//         error status
//=============================================================================
Status MemPartition::setDiskPartition(DiskPartition *dp)
{
    diskPartition = dp;

    return OK;
} // end setDiskPartition


//=============================================================================
// Method: getDiskPartition
//     Description:
//         get disk partition that contains tuples flushed from this mem part
//     Parameters:
//         none
//     Return Value:
//         the corresponding disk partition
//=============================================================================
DiskPartition* MemPartition::getDiskPartition(void)
{
    return diskPartition;

} // end getDiskPartition


//=============================================================================
// Method: insert
//     Description:
//         insert a tuple into the mem partition
//     Parameters:
//         pTuple - new tuple to insert
//     Return Value:
//         error status
//=============================================================================
Status MemPartition::insert(StampedTuple *pTuple)
{
    int chainIndex = hash (pTuple->tuple->unique1);
    pTuple->next = NULL;

    // why not just stick it at the front of the chain?
    if (tuples[chainIndex] == NULL) {
        tuples[chainIndex] = pTuple;
    } else {
        pTuple->next = tuples[chainIndex];
        tuples[chainIndex] = pTuple;
    }
    tupleCount++;
    return OK;

} // end insert


//=============================================================================
// Method: probe
//     Description:
//         probe this partition with a given tuple to look for matches
//     Parameters:
//         pTuple - tuple to probe with 
//         pMathingTuple - fill in with pointer to matching tuple if found
//     Return Value:
//         error status
//=============================================================================
Status MemPartition::probe(StampedTuple *pTuple, StampedTuple* &pMatchingTuple)
{
    pMatchingTuple = NULL;
    int chainIndex = hash (pTuple->tuple->unique1);

    StampedTuple *current = tuples[chainIndex];
    // nobody is in this chain at all, doh!
    while (current != NULL) {

        // if match
        if (current->tuple->unique1 == pTuple->tuple->unique1) {
            // make a new output tuple
            pMatchingTuple = current;
            return OK;
        }
        current = current->next;
    }

    if (NULL == pMatchingTuple)
    {
        return NOMATCH;
    }
    return OK;
} // end probe


//=============================================================================
// Method: flush
//     Description:
//         flush contents of this memory partition to disk and free its mem
//     Parameters:
//         bytesFlushed - set to number of bytes that were flushed
//     Return Value:
//         error status
//=============================================================================
Status MemPartition::flush(int &bytesFlushed)
{

    Status status = OK;
    bytesFlushed = 0;
    int writeCount = 0;

    // open our disk partition for appending
    if ((status = diskPartition->beginAppend ()) != OK) {
        cout << "error begin append" << endl;
        return status;
    }

    // march through the hash table, deleting StampedTuples as we go
    // sending each one to disk
    for (int i = 0; i < NUM_CHAINS; i++) {
        StampedTuple *current = tuples[i];
        StampedTuple *previous;
        while (current != NULL) {
            previous = current->next;
            if ((status = diskPartition->appendTuple (current)) != OK) return status;
            writeCount++;
            delete current->tuple;
            delete current;
            current = previous;
        }
        tuples[i] = NULL;
    }



    if (tupleCount != writeCount) {
        cout << "tuples written to disk did not match count in memory" << endl;
        return FILE_ERROR;
    }

    // close the disk partition
    if ((status = diskPartition->endAppend ()) != OK) {
        cout << "error end append" << endl;
        return status;
    }

    // update the number of tuples in this partition (in memory)
    bytesFlushed = tupleCount * sizeof (StampedTuple) + tupleCount * sizeof (WITuple);
    tupleCount = 0;

    return OK;
} // end flush


//=============================================================================
// Method: getSize
//     Description:
//         get the current size in bytes of the mem partition
//     Parameters:
//         none
//     Return Value:
//         number of bytes of mem currently occupied
//=============================================================================
int MemPartition::getSize(void)
{
    return tupleCount * (sizeof (StampedTuple) + sizeof (WITuple));
} // end getSize


//=============================================================================
// Method: getTupleCount
//     Description:
//         get the number of tuples currently in this mem part
//     Parameters:
//         none
//     Return Value:
//         number of tuples
//=============================================================================
int MemPartition::getTupleCount(void)
{
    return tupleCount;
} // end getTupleCount


//=============================================================================
// Method: hash
//     Description:
//         simple hash function
//     Parameters:
//         key - key to hash on 
//     Return Value:
//         hash value of key
//=============================================================================
int MemPartition::hash(int key) 
{
    return key % NUM_CHAINS;
} // end hash


//=============================================================================
// Method: operator<<
//     Description:
//         for debugging
//     Parameters:
//         -
//     Return Value:
//         -
//=============================================================================
ostream& operator << (ostream& os, MemPartition& memPartition) 
{
    os << "mem " << memPartition.id << "(" << memPartition.tupleCount << ") ";
    os << *memPartition.diskPartition;
    return os;
}
