//=============================================================================
// File: xjoin.cpp
//    
// Created Tue Apr 11 13:41:19 2000
//    
// Copyright (c) 2000 
// Justin E. Forrester (jforrest@cs.wisc.edu)
// Jonathan Ledlie (ledlie@cs.wisc.edu)
// All Rights Reserved
// 
// Description:
//    implementation of xjoin stages 
//     
//=============================================================================
#include <sys/select.h>
#include <sys/time.h>
#include "xjoin.h"
#include "tuple.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ddi.h>
#include <stdio.h>
#include <errno.h>
#include <strings.h>
#include <netdb.h>
#include "output.h"

//=============================================================================
// Method: ctor
//     Description:
//         constructor
//     Parameters:
//         max - maximum amount of memory to use (in bytes)
//     Return Value:
//         none
//=============================================================================
XJoin::XJoin(int max)
{
    srand(time(NULL));

    connfd1 = 0;
    connfd2 = 0;
    conn1done = 0;
    conn2done = 0;

    maxMem = max;
    currMem = 0;

    tuplesFlushed = 0;
    inMemoryMatchCount = 0;
    proberTupleCount = 0;

    useSecondStage = true;

    for (int i = 0; i < NUM_PARTITIONS; i++)
    {
        memPartitions1[i].setDiskPartition(&diskPartitions1[i]);
        memPartitions2[i].setDiskPartition(&diskPartitions2[i]);
        
        //================================================
        // FIXME: allocate one memory block for each partition
        //================================================
    }
    
    char filename[MAX_FILENAME];
    for (int i = 0; i < NUM_PARTITIONS; i++)
    {
        memset(filename, 0, MAX_FILENAME);
        sprintf(filename, "/tmp/dp1_%d.raw", i);
	unlink (filename);
        diskPartitions1[i].setFilename(filename);

        memset(filename, 0, MAX_FILENAME);
        sprintf(filename, "/tmp/dp2_%d.raw", i);
	unlink (filename);
        diskPartitions2[i].setFilename(filename);
    }
        
} // end ctor


//=============================================================================
// Method: dtor
//     Description:
//         destructor
//     Parameters:
//         none
//     Return Value:
//         none
//=============================================================================
XJoin::~XJoin (void) 
{
    close (connfd1);
    close (connfd2);
    close (conn1done);
    close (conn2done);
}


//=============================================================================
// Method: toggleSecondStage
//     Description:
//         toggle the use of the second stage
//     Parameters:
//         useSecondStage - if true, 2nd stage will be used
//     Return Value:
//         none
//=============================================================================
void XJoin::toggleSecondStageUse(bool useSecondStage)
{
    this->useSecondStage = useSecondStage;

} // end toggleSecondStageUse


//=============================================================================
// Method: init
//     Description:
//         start the xjoin by accepting connections from tuple sources
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status XJoin::init(void)
{
    return acceptConnections();
        
} // end init


//=============================================================================
// Method: exec
//     Description:
//         execute XJoin
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status XJoin::exec(void)
{
    Status status;

    // while the first stage is not done, do 1 & 2
    while (1)
    {
        // do the first stage
        if ((status = firstStage()) == OK)
            break;

	ASSERT(status == CONNECTIONS_BLOCKED);

	// do second stage
	if (useSecondStage)
            secondStage();
    }
    
    // finally do third stage
    if ((status = thirdStage()) != OK) {
      cout << "Third stage error: " << status << endl;
    }

    for (int i=0; i < NUM_PARTITIONS; i++) {
        diskPartitions1[i].cleanup();
        diskPartitions2[i].cleanup();
    }
    return OK;
} // end exec


//=============================================================================
// Method: acceptConnections
//     Description:
//         wait for an incoming connection attempts and accept 2 connections
//     Parameters:
//         connfd - result param that holds new connection file descriptor
//     Return Value:
//         error status
//=============================================================================
Status XJoin::acceptConnections(void)
{
    int sockfd;
    int clientLen;
    struct sockaddr_in serverAddr, clientAddr;
    
    // create socket
    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
    {
        perror("receive: could not create socket: ");
        return SOCKET_ERR;
    }

    // bind to local port
    memset(&serverAddr, 0, sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = htonl(INADDR_ANY);
    serverAddr.sin_port = htons(TUPLEPORT);
    
    if (bind(sockfd, (sockaddr *) &serverAddr, sizeof(serverAddr)) == -1)
    {
        perror("receive: error binding to receive port: ");
        close(sockfd);
        return SOCKET_ERR;
    }

    // listen for incoming tuple streams
    if (listen(sockfd, 1) == -1)
    {
        perror("error listening: ");
        close(sockfd);
        return SOCKET_ERR;
    }

    // accept the connection
    memset(&clientAddr, 0, sizeof(clientAddr));
    clientLen = 0;
    if ((connfd1 = accept(sockfd, (sockaddr *)&clientAddr, &clientLen)) == -1)
    {
        perror("error accepting connection: ");
        close(sockfd);
        return SOCKET_ERR;
    }

    // listen for incoming tuple streams
    if (listen(sockfd, 1) == -1)
    {
        perror("error listening: ");
        close(sockfd);
        return SOCKET_ERR;
    }

    // accept the connection
    memset(&clientAddr, 0, sizeof(clientAddr));
    clientLen = 0;
    if ((connfd2 = accept(sockfd, (sockaddr *)&clientAddr, &clientLen)) == -1)
    {
        perror("error accepting connection: ");
        close(sockfd);
        return SOCKET_ERR;
    }
    
    // return the new socket
    close(sockfd);
    return OK;

} // end acceptConnection


//=============================================================================
// Method: getTuple
//     Description:
//         read a tuple from a socket and process into partition
//     Parameters:
//         connfd - connection fd to read from
//         insertParitions
//         probePartitions
//     Return Value:
//         error status
//=============================================================================
Status XJoin::getTuple(int connfd, 
                       MemPartition *insertPartitions, 
                       MemPartition *probePartitions)
{
    StampedTuple *pNewTuple = new StampedTuple;
    pNewTuple->DTS.setToMaxTime();
    pNewTuple->tuple = new WITuple;
    pNewTuple->next = NULL;

    // do read
    int numRead = 0;
    if ((numRead = read(connfd, pNewTuple->tuple, sizeof(WITuple))) == -1)
    {
        perror("Error reading from socket: ");
        return SYSTEM_ERR;
    }

    if (numRead != sizeof(WITuple))
    {
      delete pNewTuple->tuple;
      delete pNewTuple;
      return SOURCE_CONNECTION_DONE;
    }
    else
    {
        // add the new tuple
        int partition = getPartition(pNewTuple->tuple);

        insertPartitions[partition].insert(pNewTuple);
        StampedTuple *pMatchingTuple;
        probePartitions[partition].probe(pNewTuple, pMatchingTuple);
                
        if (pMatchingTuple)
        {
	  inMemoryMatchCount++;

            OutputTuple *output = new OutputTuple;
            if ( ((output->tuple1.tuple = new WITuple) == 0) ||
		 ((output->tuple2.tuple = new WITuple) == 0) ) {
                perror("error allocating");
		exit (0);
	    }


            // copy the data into it
            memcpy (output->tuple1.tuple, pNewTuple->tuple, sizeof(WITuple));
            memcpy (output->tuple2.tuple, pMatchingTuple->tuple, sizeof(WITuple));
                    
            // send it to the output object
            Output::writeTuple (output, false);

	    delete output->tuple1.tuple;
	    delete output->tuple2.tuple;
	    delete output;
                    
        }
    }
    return OK;

} // end getTuple


//=============================================================================
// Method: firstStage
//     Description:
//         perform first stage functions
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status XJoin::firstStage(void)
{
    Status status;

    // find out which socket is ready
    while (!conn1done || !conn2done)
    {
        fd_set fdset;
        int maxfd = max(connfd1, connfd2) + 1;
        FD_ZERO(&fdset);
        FD_SET(connfd1, &fdset);
        FD_SET(connfd2, &fdset);
        timeval selectTimeout;
        selectTimeout.tv_sec = 0;
        selectTimeout.tv_usec = 0;
        if (select(maxfd, &fdset, NULL, NULL, &selectTimeout)  == -1)
        {
            perror("error on select");
            return SYSTEM_ERR;
        }

        // check to see if either of the conns have data ready
        if (!FD_ISSET(connfd1, &fdset) && !FD_ISSET(connfd2, &fdset))
        {
            // we are blocked on both, return status
            return CONNECTIONS_BLOCKED;
        }
        else // one of the connections is ready with another tuple
        {
            // make room for tuple if necessary
            while (maxMem - currMem < (int) (sizeof(WITuple) + sizeof(StampedTuple)))
            {
                // out of mem, free biggest partition
                MemPartition *victimPartition = &memPartitions1[0];
                for (int i = 0; i < NUM_PARTITIONS; i++)
                {
                    if (memPartitions1[i].getSize() >
                        victimPartition->getSize())
                    {
                        victimPartition = &memPartitions1[i];
                    }
                }
                for (int i = 0; i < NUM_PARTITIONS; i++)
                {
                    if (memPartitions2[i].getSize() >
                        victimPartition->getSize())
                    {
                        victimPartition = &memPartitions2[i];
                    }
                }
                // flush the victim
                int bytesFlushed = 0;

		tuplesFlushed += victimPartition->getTupleCount();

		if ((status = victimPartition->flush(bytesFlushed)) != OK)
                {
                    return status;
                }
                currMem -= bytesFlushed;
		ASSERT (bytesFlushed > 0);
            } // while there isn't enough memory for a new tuple
        }

        // read tuple from selected socket
        if (FD_ISSET(connfd1, &fdset))
        {
            status = getTuple(connfd1,
                              memPartitions1,
                              memPartitions2);
            if (OK == status)
                currMem += sizeof(WITuple);
            if (SOURCE_CONNECTION_DONE == status)
                conn1done = true;
        }

        // read tuple from selected socket
        if (FD_ISSET(connfd2, &fdset))
        {
            status = getTuple(connfd2,
                              memPartitions2,
                              memPartitions1);
            if (OK == status)
                currMem += sizeof(WITuple);
            if (SOURCE_CONNECTION_DONE == status)
                conn2done = true;
        }
    } // while not done

    // done
    return OK;

}


//=============================================================================
// Method: secondStage
//     Description:
//         perform second stage of XJoin algorithm
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status XJoin::secondStage(void)
{
    DiskPartition *proberDiskPart = NULL;
    MemPartition *probeeMemPart = NULL;

    // use round-robin scheduling for what partitions to use
    // NOTE: tried random here and it did not show significant
    // performance gains -- RR kicks butt -- JEF
    static int probePart = 0;
    static int probeRel = 0;

    if (probeRel == 0)
    {
        proberDiskPart = &diskPartitions1[probePart];
        probeeMemPart = &memPartitions2[probePart];
        probeRel = 1;
    }
    else
    {
        proberDiskPart = &diskPartitions2[probePart];
        probeeMemPart = &memPartitions1[probePart];
        probeRel = 0;
    }    

    probePart = (probePart + 1) % NUM_PARTITIONS;

    Status status = OK;
    if (proberDiskPart->getTupleCount() != 0)
    {
        // we picked a partition that has tuples, to a second stage probe
        Output::recordSecondStageRun();

        status = probeFromDisk(probeeMemPart, proberDiskPart, true);
    }

    return status;

} // end secondStage


//=============================================================================
// Method: thirdStage
//     Description:
//         run 3rd stage of XJoin algorithm
//     Parameters:
//         none
//     Return Value:
//         error status
//=============================================================================
Status XJoin::thirdStage(void)
{
    Status status = OK;

    cout << "Begin Third Stage" << endl;

    // find the smaller of the two relations
    int rel1Size = 0;
    int rel2Size = 0;

    for (int i = 0; i < NUM_PARTITIONS; i++)
    {
        rel1Size += memPartitions1[i].getTupleCount();
        rel1Size += diskPartitions1[i].getTupleCount();
        rel2Size += memPartitions2[i].getTupleCount();
        rel2Size += diskPartitions2[i].getTupleCount();
    }

    MemPartition *probeeMemPart = NULL;
    MemPartition *proberMemPart = NULL;
    DiskPartition *proberDiskPart = NULL;

    int probeeTupleCount  = 0;

    for (int i = 0; i < NUM_PARTITIONS; i++)
    {
        if (rel1Size < rel2Size)
        {
            probeeMemPart = &memPartitions1[i];
            proberMemPart = &memPartitions2[i];
            proberDiskPart = &diskPartitions2[i];
        }
        else
        {
            probeeMemPart = &memPartitions2[i];
            proberMemPart = &memPartitions1[i];
            proberDiskPart = &diskPartitions1[i];
        }

        // check to see if there are any tuples that have been flushed to disk
        if (proberDiskPart->getTupleCount() > 0)
        {

            // read all of mem(probee) part in off of disk
            if (probeeMemPart->getDiskPartition()->getTupleCount() >  0) {
                if ((status = probeeMemPart->getDiskPartition()->beginRead()) != OK) {
                    cout << "Could not begin read in third stage" << endl;
                    return status;
                }
		probeeTupleCount += probeeMemPart->getDiskPartition()->getTupleCount();
                status = OK;

		StampedTuple *stampedTuple = new StampedTuple();
		WITuple *wiTuple = new WITuple;
		stampedTuple->tuple = wiTuple;

                while (status == OK) {
   
                    if ( (status = probeeMemPart->getDiskPartition()->nextTuple(stampedTuple)) == OK) {
		      status =  probeeMemPart->insert (stampedTuple);

		      stampedTuple = new StampedTuple;
		      wiTuple = new WITuple;
		      stampedTuple->tuple = wiTuple;
                    }
                }
		delete stampedTuple;
		delete wiTuple;
		probeeMemPart->getDiskPartition()->endRead(false);
            }
            
	    status = probeFromDisk(probeeMemPart, proberDiskPart, false);
	    ASSERT(status == OK);
            status = proberMemPart->finalMemProbe(probeeMemPart);
	    ASSERT(status == OK);
        }
    }

    return OK;
} // end thirdStage


//=============================================================================
// Method: probeFromDisk
//     Description:
//         read tuples from disk partition and probe memory partition 
//         used in 2nd and 3rd stages
//     Parameters:
//         probeeMemPart - memory partition that we are going to probe
//         proberDiskPart - disk partition we are probing with
//         isSecondStage - flag indicating whether this is the 2nd stage or not
//     Return Value:
//         error status
//=============================================================================
Status XJoin::probeFromDisk(MemPartition *probeeMemPart,
                            DiskPartition *proberDiskPart,
                            bool isSecondStage)
{
    Status status = OK;
    // begin reading tuples from disk part
    StampedTuple currTuple;
    WITuple currWITuple;
    currTuple.tuple = &currWITuple; 
    StampedTuple *pMatchingTuple = NULL;

    if (proberDiskPart->getTupleCount() <  0) 
    {
        return OK;
    }

    status = proberDiskPart->beginRead();
    ASSERT(status == OK);

    while (OK == proberDiskPart->nextTuple(&currTuple))
    {

        proberTupleCount++;
 
        // check to see if they join
        if (OK == probeeMemPart->probe(&currTuple, pMatchingTuple))
        {
            // check to see if they were joined in the first stage
            if ( (currTuple.ATS > pMatchingTuple->ATS &&
                  currTuple.ATS < pMatchingTuple->DTS) ||
                 (pMatchingTuple->ATS > currTuple.ATS &&
                  pMatchingTuple->ATS < currTuple.DTS))
            {
                ;// duplicate match
            }
            else
            {
                // we did not match these in the first stage

                // check to see if we already matched these in another
                // run of the second stage
                if ((OK == proberDiskPart->checkDups(&currTuple, pMatchingTuple)) &&
                    (OK == probeeMemPart->getDiskPartition()->checkDups(pMatchingTuple, &currTuple)))
                {
                    // we have a legitimate match
                    OutputTuple *output = new OutputTuple;
                    if ((output->tuple1.tuple = new WITuple) == 0)
                        perror("error allocating");
                    output->tuple2.tuple = new WITuple;
                    
                    // copy the data into it
                    memcpy (output->tuple1.tuple, currTuple.tuple, sizeof(WITuple));
                    memcpy (output->tuple2.tuple, pMatchingTuple->tuple, sizeof(WITuple));
                    // send it to the output object
                    Output::writeTuple (output, isSecondStage);
                }
            }
        }
        pMatchingTuple = NULL;
    }

    status = proberDiskPart->endRead(isSecondStage);
    ASSERT(status == OK);
    
    return OK;
} // end probeFromDisk


//=============================================================================
// Method: getPartition
//     Description:
//         determine partition of the given tuple
//     Parameters:
//         tuple - tuple to hash
//     Return Value:
//         partition number
//=============================================================================
int XJoin::getPartition(WITuple *tuple)
{
    return tuple->unique1 % NUM_PARTITIONS;

} // end getPartition
